/////////////////////////////////////////////////////////////////////////////
// Implementation for a MFC class encapsulation of the SSL/TLS sockets

/*
 * Copyright (c) 2003 - 2005 by Daniel Weber (daniel@astolat.de)
 * All rights reserved.
 *
 * This is a SSL/TLS-implementation for Windows-Sockets in C++ written by
 * Daniel Weber (daniel@astolat.de) based an examples provided by
 * Martin Ziacek (martin.ziacek@blueyonder.co.uk).
 *
 * You are allowed to use, modify and redistribute this source-code for
 * commercial and non-commercial software as long as the following
 * conditions are met:
 * - Copyright remains bei Daniel Weber and redistributions of the source-
 *   code must retain this copyright-notice, list of conditions and
 *   disclaimer.
 * - Redistributions in binary form must mention Daniel Weber as author
 *   of parts of the SSL/TLS-implementations at program startup and/or
 *   in the documentation.
 *
 * This source-code is provided by Daniel Weber "as is" and any express or
 * implied warranties, including but not limited to the implied warranties
 * of merchantability and fitness for a particular purpose are disclaimed.
 * The authors or contributors are not liable for any damages arising out
 * of the use of this source-code, even if advised of the possibility of
 * such damage.
 */

#include "stdafx.h"

#include "SyncSocketEx.h"
#include "SecureSyncSocketEx.h"
#include "SocketException.h"

IMPLEMENT_DYNAMIC(CSecureSyncSocketEx, CSyncSocketEx)

#define IO_BUFFER_SIZE  0x10000

/////////////////////////////////////////////////////////////////////////////
// CSecureSyncSocketEx Construction

CSecureSyncSocketEx::CSecureSyncSocketEx()
{
	m_bUseSecurity = FALSE;

	m_hMyCertStore = NULL;
	m_dwProtocol = 0;
	m_bMachineStore = FALSE;
	m_hSecurity = NULL;
	m_bServer = FALSE;
	m_bAuthClient = FALSE;
	m_bConInit = FALSE;

	m_pbReceiveBuf = NULL;
	m_dwReceiveBuf = 0;

	m_bAllowPlainText = FALSE;

	m_pbIoBuffer = NULL;
	m_cbIoBuffer = 0;

	m_pCertContext = NULL;

	::memset( &m_SchannelCred, 0, sizeof( m_SchannelCred ) );

	m_hCreds.dwLower = 0;
	m_hCreds.dwUpper = 0;

	m_hContext.dwLower = 0;
	m_hContext.dwUpper = 0;

}

/* So einfach geht das hier nicht!

CSyncSocketEx::CSyncSocketEx( const CSyncSocketEx& rSrc )
{
	m_hSocket = rSrc.m_hSocket;
	m_nTimeout = rSrc.m_nTimeout;
}

void CSyncSocketEx::operator=( const CSyncSocketEx& rSrc )
{
	m_hSocket = rSrc.m_hSocket;
	m_nTimeout = rSrc.m_nTimeout;
}

*/

CSecureSyncSocketEx::~CSecureSyncSocketEx()
{
    if( m_hCreds.dwLower != 0 && m_hCreds.dwUpper != 0 )
	{
		m_SecurityFunc.FreeCredentialsHandle( &m_hCreds );
	}
	if( m_hContext.dwLower != 0 && m_hContext.dwUpper != 0 )
	{
		m_SecurityFunc.DeleteSecurityContext( &m_hContext );
	}

    if( m_pCertContext )
	{
		::CertFreeCertificateContext( m_pCertContext );
	}
    if( m_hMyCertStore )
	{
		::CertCloseStore( m_hMyCertStore, 0 );
	}
	if( m_hSecurity )
	{
		::FreeLibrary( m_hSecurity );
	}

	if( m_pbReceiveBuf )
	{
		delete [] m_pbReceiveBuf;
	}
	if( m_pbIoBuffer )
	{
		delete [] m_pbIoBuffer;
	}
}

BOOL CSecureSyncSocketEx::Create(UINT nSocketPort, int nSocketType, LPCTSTR lpszSocketAddress, BOOL bUseSecurity/* = FALSE*/, LPCTSTR pszCertName/* = NULL*/, BOOL bMachineStore/* = FALSE*/, DWORD dwProtocol/* = 0*/ )
{
	if( !CSyncSocketEx::Create( nSocketPort, nSocketType, lpszSocketAddress ) )
	{
		return FALSE;
	}

	if( bUseSecurity )
	{
		if( !LoadSecurityLibrary() )
		{
			return FALSE;
		}

		m_bUseSecurity = TRUE;
		m_strCertName = pszCertName;
		m_bMachineStore = bMachineStore;
		m_dwProtocol = dwProtocol;
	}

	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// CAscynSocket Operations

BOOL CSecureSyncSocketEx::Accept( CSecureSyncSocketEx& rConnectedSocket, SOCKADDR* lpSockAddr, int* lpSockAddrLen )
{
	if( !CSyncSocketEx::Accept( rConnectedSocket, lpSockAddr, lpSockAddrLen ) )
	{
		return FALSE;
	}

	rConnectedSocket.m_bServer = TRUE;

	if( m_bUseSecurity )
	{
		rConnectedSocket.m_bUseSecurity = TRUE;

		rConnectedSocket.m_bAuthClient = m_bAuthClient;
		rConnectedSocket.m_strCertName = m_strCertName;
		rConnectedSocket.m_dwProtocol = m_dwProtocol;
		rConnectedSocket.m_bMachineStore = m_bMachineStore;

		rConnectedSocket.LoadSecurityLibrary();
		rConnectedSocket.ServerCreateCredentials( LPCTSTR(m_strCertName), &(rConnectedSocket.m_hCreds) );
		rConnectedSocket.m_bConInit = rConnectedSocket.ServerConnect( lpSockAddr,lpSockAddrLen );
	}

	return TRUE;
}

BOOL CSecureSyncSocketEx::LateHandshake( BOOL bAuthClient, LPCTSTR pszCertName/* = NULL*/, BOOL bMachineStore/* = FALSE*/, DWORD dwProtocol/* = 0*/, LPCTSTR pszHostName/* = NULL*/ )
{
	if( !LoadSecurityLibrary() )
	{
		return FALSE;
	}

	m_bUseSecurity = TRUE;

	m_bAuthClient = bAuthClient;
	m_strCertName = pszCertName;
	m_bMachineStore = bMachineStore;
	m_dwProtocol = dwProtocol;

	if( m_bServer )
	{
		ServerCreateCredentials( LPCTSTR(m_strCertName), &m_hCreds );
		m_bConInit = ServerConnect( NULL, NULL );
	}
	else
	{
		m_bConInit = ClientConnect( pszHostName );
	}

	return m_bConInit;
}

BOOL CSecureSyncSocketEx::Close()
{
	try
	{
		if( m_bUseSecurity )
		{
			if( m_bServer )
			{
				ServerDisconect( &m_hCreds, &m_hContext );
			}
			else
			{
				ClientDisconnect( &m_hCreds, &m_hContext );
			}
		}
	}
	catch( CSocketException* e )
	{
		e->Delete();
	}

	return CSyncSocketEx::Close();
}

BOOL CSecureSyncSocketEx::Connect( LPCTSTR lpszHostAddress, UINT nHostPort )
{
	if( !CSyncSocketEx::Connect( lpszHostAddress, nHostPort ) )
	{
		return FALSE;
	}

	if( m_bUseSecurity )
	{
		m_bConInit = ClientConnect( lpszHostAddress );
		return m_bConInit;
	}
	else
	{
		return TRUE;
	}
}

BOOL CSecureSyncSocketEx::Connect( const SOCKADDR* lpSockAddr, int nSockAddrLen )
{
	if( !CSyncSocketEx::Connect( lpSockAddr, nSockAddrLen ) )
	{
		return FALSE;
	}

	if( m_bUseSecurity )
	{
		HOSTENT *pHostEnt = ::gethostbyaddr( (const char *)lpSockAddr, nSockAddrLen, AF_INET );
#ifdef _UNICODE
		DWORD dwServerName = ::MultiByteToWideChar( CP_ACP, 0, pHostEnt->h_name, -1, NULL, 0 );
		WCHAR *wszServerName = new WCHAR[dwServerName];
		::MultiByteToWideChar( CP_ACP, 0, pHostEnt->h_name, -1, wszServerName, dwServerName );
		m_bConInit = ClientConnect( wszServerName );
		delete [] wszServerName;
#else
		m_bConInit = ClientConnect( pHostEnt->h_name );
#endif
		return m_bConInit;
	}
	else
	{
		return TRUE;
	}
}

BOOL CSecureSyncSocketEx::Listen( int nConnectionBacklog, BOOL bAuthClient )
{
	m_bAuthClient = bAuthClient;
	m_bServer = TRUE;

	if( m_bUseSecurity && m_hCreds.dwLower == 0 && m_hCreds.dwUpper == 0 )
	{
		if( ServerCreateCredentials( LPCTSTR(m_strCertName), &m_hCreds ) != 0 )
		{
			return FALSE;
		}
	}
	return CSyncSocketEx::Listen( nConnectionBacklog );
}

int CSecureSyncSocketEx::Receive( void* lpBuf, int nBufLen, int nFlags )
{
	int rc = 0;

	if( m_bUseSecurity && m_bConInit )
	{
		SecPkgContext_StreamSizes Sizes;
		SECURITY_STATUS scRet;
		DWORD cbIoBufferLength;
		DWORD cbData;
		SecBufferDesc Message;
		SecBuffer Buffers[4];
	    SecBuffer* pDataBuffer;
		SecBuffer ExtraBuffer;
	    SecBuffer* pExtraBuffer;

		BYTE *pDataBuf = NULL;
		DWORD dwDataLn = 0;
		DWORD dwBufDataLn = 0;
		BOOL bCont = TRUE;

		if( m_dwReceiveBuf )
		{
			if( (DWORD)nBufLen < m_dwReceiveBuf )
			{
				rc = nBufLen;
				::memcpy( lpBuf, m_pbReceiveBuf, rc );
				::memmove( m_pbReceiveBuf, m_pbReceiveBuf + rc, m_dwReceiveBuf - rc );
				m_dwReceiveBuf -= rc;
			}
			else
			{
				rc = m_dwReceiveBuf;
				::memcpy( lpBuf, m_pbReceiveBuf, rc );
				delete [] m_pbReceiveBuf;
				m_pbReceiveBuf = NULL;
				m_dwReceiveBuf = 0;
			}
		}
		else
		{
			do
			{
				scRet = m_SecurityFunc.QueryContextAttributes( &m_hContext, SECPKG_ATTR_STREAM_SIZES, &Sizes );
				if(	scRet != SEC_E_OK )
				{
					SetLastError( scRet );
					break;
				}

				cbIoBufferLength = Sizes.cbHeader + Sizes.cbMaximumMessage + Sizes.cbTrailer;

				if( !m_pbIoBuffer )
				{
					m_pbIoBuffer = new BYTE[cbIoBufferLength];
				}
				pDataBuf = new BYTE[cbIoBufferLength];
				dwBufDataLn = cbIoBufferLength;
				if( m_pbIoBuffer == NULL || pDataBuf == NULL )
				{
					SetLastError( ERROR_OUTOFMEMORY );
					break;
				}

				do
				{
					cbData = CSyncSocketEx::Receive( m_pbIoBuffer + m_cbIoBuffer, cbIoBufferLength - m_cbIoBuffer );
					if( cbData == SOCKET_ERROR )
					{
						SetLastError( ::WSAGetLastError() );
						break;
					}
					else if( cbData == 0 )
					{
						if( m_cbIoBuffer )
						{
							TRACE( _T("**** Server unexpectedly disconnected\n") );
							scRet = SEC_E_INTERNAL_ERROR;
							break;
						}
						else
						{
							break;
						}
					}
					else
					{
						m_cbIoBuffer += cbData;
					}
//_ftprintf( stderr, _T("Receive-Ciphertext:\n") );
//for( unsigned dump = 0; dump < cbData; dump++ )
//{
//	_ftprintf( stderr, _T("%02X "), m_pbIoBuffer[dump] );
//}
//_ftprintf( stderr, _T("\n") );

					Buffers[0].pvBuffer     = m_pbIoBuffer;
					Buffers[0].cbBuffer     = m_cbIoBuffer;
					Buffers[0].BufferType   = SECBUFFER_DATA;

					Buffers[1].BufferType   = SECBUFFER_EMPTY;
					Buffers[2].BufferType   = SECBUFFER_EMPTY;
					Buffers[3].BufferType   = SECBUFFER_EMPTY;

					Message.ulVersion       = SECBUFFER_VERSION;
					Message.cBuffers        = 4;
					Message.pBuffers        = Buffers;

					scRet = m_SecurityFunc.DecryptMessage( &m_hContext, &Message, 0, NULL );
					if( scRet == SEC_E_INCOMPLETE_MESSAGE )
					{
						continue;
					}
					else if( scRet == SEC_I_CONTEXT_EXPIRED )
					{
						SetLastError(scRet);
						break;
					}
					else if( scRet != SEC_E_OK && scRet != SEC_I_RENEGOTIATE && scRet != SEC_I_CONTEXT_EXPIRED )
					{
						SetLastError(scRet);
						break;
					}

					pDataBuffer  = NULL;
					pExtraBuffer = NULL;
					for( int i = 1; i < 4; i++ )
					{
						if( pDataBuffer == NULL && Buffers[i].BufferType == SECBUFFER_DATA )
						{
							pDataBuffer = &Buffers[i];
						}
						if( pExtraBuffer == NULL && Buffers[i].BufferType == SECBUFFER_EXTRA )
						{
							pExtraBuffer = &Buffers[i];
						}
					}

					if( pDataBuffer )
					{
//_ftprintf( stderr, _T("Receive-Plaintext:\n%s\n"), CString( (LPCTSTR)(pDataBuffer->pvBuffer), pDataBuffer->cbBuffer ) );
						if( (dwDataLn + (pDataBuffer->cbBuffer)) > dwBufDataLn )
						{
							BYTE *bNewDataBuf = new BYTE[dwBufDataLn+(pDataBuffer->cbBuffer)];
							::memcpy( bNewDataBuf, pDataBuf, dwDataLn );
							delete [] pDataBuf;
							pDataBuf = bNewDataBuf;
							dwBufDataLn = dwBufDataLn + (pDataBuffer->cbBuffer);
						}
						::memcpy( pDataBuf+dwDataLn, pDataBuffer->pvBuffer, pDataBuffer->cbBuffer );
						dwDataLn += pDataBuffer->cbBuffer;
					}

					if( pExtraBuffer )
					{
						::memmove( m_pbIoBuffer, pExtraBuffer->pvBuffer, pExtraBuffer->cbBuffer );
						m_cbIoBuffer = pExtraBuffer->cbBuffer;
						continue;
					}
					else
					{
						m_cbIoBuffer = 0;
						bCont = FALSE;
					}

					if( scRet == SEC_I_RENEGOTIATE )
					{
						scRet = ClientHandshakeLoop( &m_hCreds, &m_hContext, FALSE, &ExtraBuffer );
						if( scRet != SEC_E_OK )
						{
							break;
						}

						if( ExtraBuffer.pvBuffer )
						{
							::memmove( m_pbIoBuffer, ExtraBuffer.pvBuffer, ExtraBuffer.cbBuffer );
							m_cbIoBuffer = ExtraBuffer.cbBuffer;
						}

						if( ExtraBuffer.pvBuffer )
						{
							delete [] ExtraBuffer.pvBuffer;
						}
					}
				}
				while( bCont );

			}
			while( FALSE );

			if( dwDataLn )
			{
				if( dwDataLn > (DWORD)nBufLen )
				{
					m_dwReceiveBuf = dwDataLn - ((DWORD)(nBufLen));
					m_pbReceiveBuf = new BYTE[m_dwReceiveBuf];

					::memcpy(lpBuf,pDataBuf,nBufLen);
					rc = nBufLen;

					::memcpy(m_pbReceiveBuf,pDataBuf+nBufLen,m_dwReceiveBuf);

				}
				else
				{
					::memcpy(lpBuf,pDataBuf,dwDataLn);
					rc = dwDataLn;
				}
			}

			if( pDataBuf )
			{
				delete [] pDataBuf;
			}
		}
	}
	else
	{
		if( !m_bUseSecurity || m_bAllowPlainText )
		{
			rc = CSyncSocketEx::Receive( lpBuf, nBufLen, nFlags );
		}
	}

	return rc;
}

int CSecureSyncSocketEx::Send( const void* lpBuf, int nBufLen, int nFlags )
{
	int rc = 0;
	DWORD dwTotSent = 0;

	if( m_bUseSecurity && m_bConInit )
	{
		SecPkgContext_StreamSizes Sizes;
		SECURITY_STATUS scRet;
		SecBufferDesc Message;
		SecBuffer Buffers[4];
		SecBuffer* pDataBuffer;
		SecBuffer* pExtraBuffer;

		PBYTE pbIoBuffer = NULL;
		DWORD cbIoBufferLength;
		PBYTE pbMessage;
		DWORD cbMessage;

		DWORD dwAvaLn = 0;
		DWORD dwDataToSend = 0;
		DWORD dwSendInd = 0;
		DWORD dwCurrLn = 0;

		do
		{
			scRet = m_SecurityFunc.QueryContextAttributes( &m_hContext, SECPKG_ATTR_STREAM_SIZES, &Sizes );
			if( scRet != SEC_E_OK )
			{
				SetLastError(scRet);
				break;
			}

			cbIoBufferLength = Sizes.cbHeader + Sizes.cbMaximumMessage + Sizes.cbTrailer;

			pbIoBuffer = new BYTE[cbIoBufferLength];
			if( pbIoBuffer == NULL )
			{
				SetLastError(ERROR_OUTOFMEMORY);
				break;
			}

			pbMessage = pbIoBuffer + Sizes.cbHeader;
			dwAvaLn = Sizes.cbMaximumMessage;
			dwDataToSend = ((DWORD)nBufLen);

			do
			{
				::memset( pbIoBuffer, 0, cbIoBufferLength );
				dwCurrLn = ((DWORD)nBufLen)-dwSendInd > dwAvaLn ? dwAvaLn : (((DWORD)nBufLen)-dwSendInd);

				::memcpy(pbMessage,((BYTE*)lpBuf)+dwSendInd,dwCurrLn);

//_ftprintf( stderr, _T("Send-Plaintext:\n%s\n"), CString( (LPCTSTR)pbMessage, dwCurrLn ) );

				dwSendInd += dwCurrLn;
				dwDataToSend -= dwCurrLn;

				cbMessage = dwCurrLn;

				Buffers[0].pvBuffer = pbIoBuffer;
				Buffers[0].cbBuffer = Sizes.cbHeader;
				Buffers[0].BufferType = SECBUFFER_STREAM_HEADER;

				Buffers[1].pvBuffer = pbMessage;
				Buffers[1].cbBuffer = cbMessage;
				Buffers[1].BufferType = SECBUFFER_DATA;

				Buffers[2].pvBuffer = pbMessage + cbMessage;
				Buffers[2].cbBuffer = Sizes.cbTrailer;
				Buffers[2].BufferType = SECBUFFER_STREAM_TRAILER;

				Buffers[3].BufferType = SECBUFFER_EMPTY;

				Message.ulVersion = SECBUFFER_VERSION;
				Message.cBuffers = 4;
				Message.pBuffers = Buffers;

				scRet = m_SecurityFunc.EncryptMessage( &m_hContext, 0, &Message, 0 );
				if( FAILED(scRet) )
				{
					SetLastError(scRet);
					break;
				}

				pDataBuffer = NULL;
				pExtraBuffer = NULL;
				for( int i = 1; i < 4; i++ )
				{
					if( pDataBuffer == NULL && Buffers[i].BufferType == SECBUFFER_DATA )
					{
						pDataBuffer = &Buffers[i];
					}
					if( pExtraBuffer == NULL && Buffers[i].BufferType == SECBUFFER_EXTRA )
					{
						pExtraBuffer = &Buffers[i];
					}
				}

//_ftprintf( stderr, _T("Send-Ciphertext:\n") );
//for( unsigned dump = 0; dump < Buffers[0].cbBuffer + Buffers[1].cbBuffer + Buffers[2].cbBuffer; dump++ )
//{
//	_ftprintf( stderr, _T("%02X "), pbIoBuffer[dump] );
//}
//_ftprintf( stderr, _T("\n") );

				rc = CSyncSocketEx::Send( pbIoBuffer, Buffers[0].cbBuffer + Buffers[1].cbBuffer + Buffers[2].cbBuffer, nFlags );
				if( rc == SOCKET_ERROR && WSAGetLastError() == WSAEWOULDBLOCK )
				{
					rc = nBufLen;
				}
				else
				{
					if( rc == SOCKET_ERROR )
					{
						SetLastError(WSAGetLastError());
						dwTotSent = rc;
						break;
					}
					else
					{
						dwTotSent += rc;
					}
				}

			}
			while( dwDataToSend != 0 );

		}
		while( FALSE );

		if( pbIoBuffer )
		{
			delete [] pbIoBuffer;
		}
	}
	else
	{
		if( !m_bUseSecurity || m_bAllowPlainText )
		{
			rc = CSyncSocketEx::Send( lpBuf, nBufLen, nFlags );
			if( rc == SOCKET_ERROR && GetLastError() == WSAEWOULDBLOCK )
			{
				rc = nBufLen;
			}
			dwTotSent = rc;
		}
	}

	return dwTotSent > ((DWORD)nBufLen) ? nBufLen : dwTotSent;	
}

/////////////////////////////////////////////////////////////////////////////
// CSecureSyncSocketEx Implementation

BOOL CSecureSyncSocketEx::LoadSecurityLibrary()
{
	m_hSecurity = ::LoadLibrary(_T("SCHANNEL.DLL"));
	if( m_hSecurity == NULL )
	{
		return FALSE;
	}

	INIT_SECURITY_INTERFACE pInitSecurityInterface = (INIT_SECURITY_INTERFACE)GetProcAddress( m_hSecurity, SECURITY_ENTRYPOINT_ANSI );
	if( pInitSecurityInterface == NULL )
	{
		return FALSE;
	}

	PSecurityFunctionTable pSecurityFunc = pInitSecurityInterface();
	if( pSecurityFunc == NULL )
	{
		return FALSE;
	}

	::memcpy( &m_SecurityFunc, pSecurityFunc, sizeof( m_SecurityFunc ) );

    return TRUE;
}

SECURITY_STATUS CSecureSyncSocketEx::ClientCreateCredentials( LPCTSTR pszUserName, PCredHandle phCreds )
{
	if( m_hMyCertStore == NULL )
	{
		m_hMyCertStore = ::CertOpenSystemStore( NULL, _T("MY") );
		if( !m_hMyCertStore )
		{
			return SEC_E_NO_CREDENTIALS;
		}
	}

	if( pszUserName && _tcslen(pszUserName) != 0 )
	{
		CERT_RDN_ATTR cert_rdn_attr;
		cert_rdn_attr.pszObjId = szOID_COMMON_NAME;
		cert_rdn_attr.dwValueType = CERT_RDN_ANY_TYPE;
		cert_rdn_attr.Value.cbData = _tcslen(pszUserName);

		CERT_RDN cert_rdn;
		cert_rdn.cRDNAttr = 1;
		cert_rdn.rgRDNAttr = &cert_rdn_attr;

#ifdef _UNICODE
		char *pszUn = new char[wcslen(pszUserName)+1];
		::WideCharToMultiByte(CP_ACP,0,pszUserName,-1,pszUn,wcslen(pszUserName)+1,NULL,NULL);
		cert_rdn_attr.Value.pbData = (BYTE *)pszUn;
		m_pCertContext = ::CertFindCertificateInStore( m_hMyCertStore, X509_ASN_ENCODING | PKCS_7_ASN_ENCODING, 0, CERT_FIND_SUBJECT_ATTR, &cert_rdn, NULL );
		delete [] pszUn;
#else
		cert_rdn_attr.Value.pbData = (BYTE *)pszUserName;
		m_pCertContext = ::CertFindCertificateInStore( m_hMyCertStore, X509_ASN_ENCODING | PKCS_7_ASN_ENCODING, 0, CERT_FIND_SUBJECT_ATTR, &cert_rdn, NULL );
#endif			

		if( m_pCertContext == NULL )
		{
_tprintf( _T("m_pCertContext=0x%X"), ::GetLastError() );
			return SEC_E_NO_CREDENTIALS;
		}
	}

	::memset( &m_SchannelCred, 0, sizeof( m_SchannelCred ) );
	m_SchannelCred.dwVersion  = SCHANNEL_CRED_VERSION;

	if( m_pCertContext )
	{
		m_SchannelCred.cCreds = 1;
		m_SchannelCred.paCred = &m_pCertContext;
		m_SchannelCred.hRootStore = m_hMyCertStore;
	}

	m_SchannelCred.dwMinimumCipherStrength = 80;
	m_SchannelCred.grbitEnabledProtocols = m_dwProtocol;
	m_SchannelCred.dwFlags |= SCH_CRED_NO_DEFAULT_CREDS;
//	m_SchannelCred.dwFlags |= SCH_CRED_NO_DEFAULT_CREDS | SCH_CRED_NO_SYSTEM_MAPPER | SCH_CRED_REVOCATION_CHECK_CHAIN;

	TimeStamp tsExpiry;
	return m_SecurityFunc.AcquireCredentialsHandle( NULL, UNISP_NAME, SECPKG_CRED_OUTBOUND, NULL, &m_SchannelCred, NULL, NULL, phCreds, &tsExpiry );
}

SECURITY_STATUS CSecureSyncSocketEx::ServerCreateCredentials( LPCTSTR pszUserName, PCredHandle phCreds )
{
	if( pszUserName == NULL || _tcslen(pszUserName) == 0 )
	{
		return SEC_E_NO_CREDENTIALS;
	}

	if( m_hMyCertStore == NULL )
	{
		if( m_bMachineStore )
		{
			m_hMyCertStore = ::CertOpenStore( CERT_STORE_PROV_SYSTEM, X509_ASN_ENCODING | PKCS_7_ASN_ENCODING, NULL, CERT_SYSTEM_STORE_LOCAL_MACHINE | CERT_STORE_READONLY_FLAG | CERT_STORE_OPEN_EXISTING_FLAG, L"MY" );
//			m_hMyCertStore = ::CertOpenStore( CERT_STORE_PROV_SYSTEM, X509_ASN_ENCODING | PKCS_7_ASN_ENCODING, NULL, CERT_SYSTEM_STORE_SERVICES | CERT_STORE_READONLY_FLAG | CERT_STORE_OPEN_EXISTING_FLAG, L"SMTPd\\MY" );
		}
		else
		{
			m_hMyCertStore = ::CertOpenSystemStore( NULL, _T("MY") );
		}

		if( !m_hMyCertStore )
		{
			return SEC_E_NO_CREDENTIALS;
		}
	}

	CERT_RDN_ATTR cert_rdn_attr;
	cert_rdn_attr.pszObjId = szOID_COMMON_NAME;
	cert_rdn_attr.dwValueType = CERT_RDN_ANY_TYPE;
	cert_rdn_attr.Value.cbData = _tcslen(pszUserName);

	CERT_RDN cert_rdn;
	cert_rdn.cRDNAttr = 1;
	cert_rdn.rgRDNAttr = &cert_rdn_attr;

#ifdef _UNICODE
	char *pszUn = new char[wcslen(pszUserName)+1];
	::WideCharToMultiByte( CP_ACP, 0, pszUserName, -1, pszUn, wcslen(pszUserName)+1, NULL, NULL );
	cert_rdn_attr.Value.pbData = (BYTE*)pszUn;
	m_pCertContext = ::CertFindCertificateInStore( m_hMyCertStore, X509_ASN_ENCODING | PKCS_7_ASN_ENCODING, 0, CERT_FIND_SUBJECT_ATTR, &cert_rdn, NULL );
	delete [] pszUn;
#else
	cert_rdn_attr.Value.pbData = (BYTE*)pszUserName;
	m_pCertContext = ::CertFindCertificateInStore( m_hMyCertStore, X509_ASN_ENCODING | PKCS_7_ASN_ENCODING, 0, CERT_FIND_SUBJECT_ATTR, &cert_rdn, NULL );
#endif

	if( m_pCertContext == NULL )
	{
_tprintf( _T("m_pCertContext=0x%X"), ::GetLastError() );
		return SEC_E_NO_CREDENTIALS;
	}

	::memset( &m_SchannelCred, 0, sizeof(m_SchannelCred) );
	m_SchannelCred.dwVersion = SCHANNEL_CRED_VERSION;
	m_SchannelCred.cCreds = 1;
	m_SchannelCred.paCred = &m_pCertContext;
	m_SchannelCred.hRootStore = m_hMyCertStore;
	m_SchannelCred.dwMinimumCipherStrength = 80;
	m_SchannelCred.grbitEnabledProtocols = m_dwProtocol;
//	m_SchannelCred.dwFlags |= SCH_CRED_NO_DEFAULT_CREDS;
	m_SchannelCred.dwFlags |= SCH_CRED_NO_DEFAULT_CREDS | SCH_CRED_NO_SYSTEM_MAPPER | SCH_CRED_REVOCATION_CHECK_CHAIN;
		
	TimeStamp tsExpiry;
	SECURITY_STATUS ss = m_SecurityFunc.AcquireCredentialsHandle( NULL, UNISP_NAME, SECPKG_CRED_INBOUND, NULL, &m_SchannelCred, NULL, NULL, phCreds, &tsExpiry );
	if( ss != 0 )
	{
_tprintf( _T("AcquireCredentialsHandle=0x%X"), ss );
	}
	return ss;
}

BOOL CSecureSyncSocketEx::ClientConnect( LPCTSTR szHostName )
{
	if( ClientCreateCredentials( LPCTSTR(m_strCertName), &m_hCreds ) )
	{
		return FALSE;
	}

	SecBuffer ExtraData;
	if( ClientHandshake( &m_hCreds, szHostName, &m_hContext, &ExtraData ) )
	{
		return FALSE;
	}

	PCCERT_CONTEXT pRemoteCertContext = NULL;
	SECURITY_STATUS Status = m_SecurityFunc.QueryContextAttributes( &m_hContext, SECPKG_ATTR_REMOTE_CERT_CONTEXT, (PVOID)&pRemoteCertContext );
	if( Status != SEC_E_OK )
	{
		return FALSE;
	}
/*
	Status = ClientVerifyCertificate( pRemoteCertContext, szHostName, 0 );
	if( Status )
	{
		return FALSE;
	}
*/
	::CertFreeCertificateContext( pRemoteCertContext );	//pRemoteCertContext

	return TRUE;
}

SECURITY_STATUS CSecureSyncSocketEx::ClientHandshake( PCredHandle phCreds, LPCTSTR pszServerName, CtxtHandle *phContext, SecBuffer *pExtraData )
{
    TimeStamp tsExpiry;
    SecBufferDesc OutBuffer;
    SecBuffer OutBuffers[1];

    DWORD dwSSPIFlags = ISC_REQ_SEQUENCE_DETECT | ISC_REQ_REPLAY_DETECT | ISC_REQ_CONFIDENTIALITY | ISC_RET_EXTENDED_ERROR | ISC_REQ_ALLOCATE_MEMORY | ISC_REQ_STREAM;
    DWORD dwSSPIOutFlags;

    OutBuffers[0].pvBuffer   = NULL;
    OutBuffers[0].BufferType = SECBUFFER_TOKEN;
    OutBuffers[0].cbBuffer   = 0;

    OutBuffer.cBuffers = 1;
    OutBuffer.pBuffers = OutBuffers;
    OutBuffer.ulVersion = SECBUFFER_VERSION;

	SECURITY_STATUS scRet = m_SecurityFunc.InitializeSecurityContext( phCreds, NULL, (LPTSTR)pszServerName, dwSSPIFlags, 0, /*SECURITY_NATIVE_DREP*/0, NULL, 0, phContext, &OutBuffer, &dwSSPIOutFlags, &tsExpiry );
    if( scRet != SEC_I_CONTINUE_NEEDED )
	{
        return scRet;
    }

    if( OutBuffers[0].cbBuffer != 0 && OutBuffers[0].pvBuffer != NULL )
	{
		m_bAllowPlainText = TRUE;
		DWORD cbData = Send(OutBuffers[0].pvBuffer,OutBuffers[0].cbBuffer);
		m_bAllowPlainText = FALSE;

        if( cbData == SOCKET_ERROR || cbData == 0 )
		{
            m_SecurityFunc.FreeContextBuffer(OutBuffers[0].pvBuffer);
            m_SecurityFunc.DeleteSecurityContext(phContext);
            return SEC_E_INTERNAL_ERROR;
        }

        m_SecurityFunc.FreeContextBuffer(OutBuffers[0].pvBuffer);
        OutBuffers[0].pvBuffer = NULL;
    }

    SECURITY_STATUS rc = ClientHandshakeLoop( phCreds, phContext, TRUE, pExtraData );

	if( pExtraData->pvBuffer )
		delete [] pExtraData->pvBuffer;

	return rc;
}

SECURITY_STATUS CSecureSyncSocketEx::ClientHandshakeLoop( PCredHandle phCreds, CtxtHandle *phContext, BOOL fDoInitialRead, SecBuffer *pExtraData )
{
    TimeStamp tsExpiry;
    SecBufferDesc InBuffer;
    SecBuffer InBuffers[2];
    SecBufferDesc OutBuffer;
    SecBuffer OutBuffers[1];

//	DWORD dwSSPIFlags = ISC_REQ_SEQUENCE_DETECT | ISC_REQ_REPLAY_DETECT | ISC_REQ_CONFIDENTIALITY | ISC_RET_EXTENDED_ERROR | ISC_REQ_ALLOCATE_MEMORY | ISC_REQ_STREAM | ISC_REQ_MANUAL_CRED_VALIDATION;

	// Check server certificate manually (expired, wrong name, unknown authority, withdrawn, ...)
	DWORD dwSSPIFlags = ISC_REQ_SEQUENCE_DETECT | ISC_REQ_REPLAY_DETECT | ISC_REQ_CONFIDENTIALITY | ISC_RET_EXTENDED_ERROR | ISC_REQ_ALLOCATE_MEMORY | ISC_REQ_STREAM | ISC_REQ_MANUAL_CRED_VALIDATION;
    DWORD dwSSPIOutFlags;
    DWORD cbData;

    PUCHAR IoBuffer = new BYTE[IO_BUFFER_SIZE];
    if( IoBuffer == NULL )
	{
        return SEC_E_INTERNAL_ERROR;
    }
    DWORD cbIoBuffer = 0;

    BOOL fDoRead = fDoInitialRead;

    SECURITY_STATUS scRet = SEC_I_CONTINUE_NEEDED;
    while( scRet == SEC_I_CONTINUE_NEEDED || scRet == SEC_E_INCOMPLETE_MESSAGE || scRet == SEC_I_INCOMPLETE_CREDENTIALS )
	{
        if( 0 == cbIoBuffer || scRet == SEC_E_INCOMPLETE_MESSAGE )
		{
            if( fDoRead )
			{
				m_bAllowPlainText = TRUE;
				cbData = Receive( (LPSTR)(IoBuffer + cbIoBuffer), IO_BUFFER_SIZE - cbIoBuffer );
				m_bAllowPlainText = FALSE;

                if( cbData == SOCKET_ERROR )
				{
					scRet = SEC_E_INTERNAL_ERROR;
					break;
                }
				else if( cbData == 0 )
				{
                    scRet = SEC_E_INTERNAL_ERROR;
                    break;
                }
                cbIoBuffer += cbData;
            }
			else
			{
                fDoRead = TRUE;
            }
        }

        InBuffers[0].pvBuffer = IoBuffer;
        InBuffers[0].cbBuffer = cbIoBuffer;
        InBuffers[0].BufferType = SECBUFFER_TOKEN;

        InBuffers[1].pvBuffer = NULL;
        InBuffers[1].cbBuffer = 0;
        InBuffers[1].BufferType = SECBUFFER_EMPTY;

        InBuffer.cBuffers = 2;
        InBuffer.pBuffers = InBuffers;
        InBuffer.ulVersion = SECBUFFER_VERSION;

        OutBuffers[0].pvBuffer = NULL;
        OutBuffers[0].BufferType = SECBUFFER_TOKEN;
        OutBuffers[0].cbBuffer = 0;

        OutBuffer.cBuffers = 1;
        OutBuffer.pBuffers = OutBuffers;
        OutBuffer.ulVersion = SECBUFFER_VERSION;

        scRet = m_SecurityFunc.InitializeSecurityContext( phCreds, phContext, NULL, dwSSPIFlags, 0, SECURITY_NATIVE_DREP, &InBuffer, 0, NULL, &OutBuffer, &dwSSPIOutFlags, &tsExpiry );
        if( scRet == SEC_E_OK || scRet == SEC_I_CONTINUE_NEEDED || FAILED(scRet) && ( dwSSPIOutFlags & ISC_RET_EXTENDED_ERROR ) )
		{
            if( OutBuffers[0].cbBuffer != 0 && OutBuffers[0].pvBuffer != NULL )
			{
				m_bAllowPlainText = TRUE;
				cbData = Send( (LPCSTR)(OutBuffers[0].pvBuffer), OutBuffers[0].cbBuffer );
				m_bAllowPlainText = FALSE;

                if( cbData == SOCKET_ERROR || cbData == 0 )
				{
                    m_SecurityFunc.FreeContextBuffer(OutBuffers[0].pvBuffer);
                    m_SecurityFunc.DeleteSecurityContext(phContext);
                    return SEC_E_INTERNAL_ERROR;
                }

                m_SecurityFunc.FreeContextBuffer(OutBuffers[0].pvBuffer);
                OutBuffers[0].pvBuffer = NULL;
            }
        }

        if( scRet == SEC_E_INCOMPLETE_MESSAGE )
		{
            continue;
        }

        if( scRet == SEC_E_OK )
		{
            if( InBuffers[1].BufferType == SECBUFFER_EXTRA )
			{
                pExtraData->pvBuffer = new BYTE[InBuffers[1].cbBuffer];

                if( pExtraData->pvBuffer == NULL )
				{
                    return SEC_E_INTERNAL_ERROR;
                }

				::memmove( pExtraData->pvBuffer, IoBuffer + (cbIoBuffer - InBuffers[1].cbBuffer ), InBuffers[1].cbBuffer );

                pExtraData->cbBuffer = InBuffers[1].cbBuffer;
                pExtraData->BufferType = SECBUFFER_TOKEN;
            }
			else
			{
                pExtraData->pvBuffer = NULL;
                pExtraData->cbBuffer = 0;
                pExtraData->BufferType = SECBUFFER_EMPTY;
            }

            break;
        }

        if( FAILED(scRet) )
		{
            break;
        }

        if( scRet == SEC_I_INCOMPLETE_CREDENTIALS )
		{
			break;
        }

        if( InBuffers[1].BufferType == SECBUFFER_EXTRA )
		{
			::memmove( IoBuffer, IoBuffer + (cbIoBuffer - InBuffers[1].cbBuffer ), InBuffers[1].cbBuffer );
            cbIoBuffer = InBuffers[1].cbBuffer;
        }
		else
		{
            cbIoBuffer = 0;
        }
    }

    if( FAILED(scRet) )
	{
		m_SecurityFunc.DeleteSecurityContext(phContext);
    }

	if( IoBuffer )
		delete [] IoBuffer;

    return scRet;
}

DWORD CSecureSyncSocketEx::ClientVerifyCertificate( PCCERT_CONTEXT pServerCert, LPCTSTR pszServerName, DWORD dwCertFlags )
{
    HTTPSPolicyCallbackData polHttps;
    CERT_CHAIN_POLICY_PARA PolicyPara;
    CERT_CHAIN_POLICY_STATUS PolicyStatus;
    CERT_CHAIN_PARA ChainPara;
    PCCERT_CHAIN_CONTEXT pChainContext = NULL;

    DWORD Status;
    PWSTR pwszServerName = NULL;
#ifndef _UNICODE
    DWORD cchServerName;
#endif

	do
	{
		if( pServerCert == NULL )
		{
			Status = (DWORD)SEC_E_WRONG_PRINCIPAL;
			break;
		}
		int iRc = ::CertVerifyTimeValidity( NULL, pServerCert->pCertInfo );
		if( iRc != 0 )
		{
			Status = (DWORD)SEC_E_CERT_EXPIRED;
			break;
		}

#ifndef _UNICODE
		if( pszServerName == NULL || strlen(pszServerName) == 0 )
		{
			return (DWORD)SEC_E_WRONG_PRINCIPAL;
		}

		cchServerName = ::MultiByteToWideChar( CP_ACP, 0, pszServerName, -1, NULL, 0 );
		pwszServerName = new WCHAR[cchServerName];
		if( pwszServerName == NULL )
		{
			return (DWORD)SEC_E_INSUFFICIENT_MEMORY;
		}
		cchServerName = ::MultiByteToWideChar( CP_ACP, 0, pszServerName, -1, pwszServerName, cchServerName );
		if( cchServerName == 0 )
		{
			return (DWORD)SEC_E_WRONG_PRINCIPAL;
		}
#else
		pwszServerName = (WCHAR*)pszServerName;
#endif

		::memset( &ChainPara, 0, sizeof( ChainPara ) );
		ChainPara.cbSize = sizeof( ChainPara );

		if( !::CertGetCertificateChain( NULL, pServerCert, NULL, NULL, &ChainPara, 0, NULL, &pChainContext ) )
		{
			Status = ::GetLastError();
			break;
		}

		::memset( &polHttps, 0, sizeof( HTTPSPolicyCallbackData ) );
		polHttps.cbStruct = sizeof( HTTPSPolicyCallbackData );
		polHttps.dwAuthType = AUTHTYPE_SERVER;
		polHttps.fdwChecks = dwCertFlags;
		polHttps.pwszServerName = pwszServerName;

		::memset( &PolicyPara, 0, sizeof( PolicyPara ) );
		PolicyPara.cbSize = sizeof( PolicyPara );
		PolicyPara.pvExtraPolicyPara = &polHttps;

		::memset( &PolicyStatus, 0, sizeof( PolicyStatus ) );
		PolicyStatus.cbSize = sizeof(PolicyStatus);

		if( !::CertVerifyCertificateChainPolicy( CERT_CHAIN_POLICY_SSL, pChainContext, &PolicyPara, &PolicyStatus ) )
		{
			Status = ::GetLastError();
			break;
		}

		if( PolicyStatus.dwError )
		{
			Status = PolicyStatus.dwError;
			break;
		}

		PCERT_CONTEXT *pCerts = new PCERT_CONTEXT[pChainContext->cChain];

		for( DWORD i = 0; i < pChainContext->cChain; i++ )
		{
			pCerts[i] = (PCERT_CONTEXT)(pChainContext->rgpChain[i]->rgpElement[0]->pCertContext);
		}
		
		CERT_REVOCATION_STATUS revStat;
		revStat.cbSize = sizeof( CERT_REVOCATION_STATUS );

		BOOL bRc = ::CertVerifyRevocation( X509_ASN_ENCODING | PKCS_7_ASN_ENCODING, CERT_CONTEXT_REVOCATION_TYPE, pChainContext->cChain, (void **)pCerts, CERT_VERIFY_REV_CHAIN_FLAG, NULL, &revStat );
		if( !bRc )
		{
			Status = ::GetLastError();
			break;
		}

		delete [] pCerts;

		Status = SEC_E_OK;

	}
	while (FALSE);

    if( pChainContext )
	{
        ::CertFreeCertificateChain(pChainContext);
    }

#ifndef _UNICODE
	if( pwszServerName )
		delete [] pwszServerName;
#endif

    return Status;
}

LONG CSecureSyncSocketEx::ClientDisconnect( PCredHandle phCreds, CtxtHandle *phContext )
{
    DWORD cbData;

    SecBufferDesc OutBuffer;
    SecBuffer OutBuffers[1];

	DWORD dwType = SCHANNEL_SHUTDOWN;

    OutBuffers[0].pvBuffer = &dwType;
    OutBuffers[0].BufferType = SECBUFFER_TOKEN;
    OutBuffers[0].cbBuffer = sizeof(dwType);

    OutBuffer.cBuffers = 1;
    OutBuffer.pBuffers = OutBuffers;
    OutBuffer.ulVersion = SECBUFFER_VERSION;

	DWORD Status = m_SecurityFunc.ApplyControlToken( phContext, &OutBuffer );
	if( FAILED(Status) )
	{
		m_SecurityFunc.DeleteSecurityContext(phContext);
	    return Status;
	}

	DWORD dwSSPIFlags = ISC_REQ_SEQUENCE_DETECT | ISC_REQ_REPLAY_DETECT | ISC_REQ_CONFIDENTIALITY | ISC_RET_EXTENDED_ERROR | ISC_REQ_ALLOCATE_MEMORY | ISC_REQ_STREAM;

	OutBuffers[0].pvBuffer = NULL;
	OutBuffers[0].BufferType = SECBUFFER_TOKEN;
	OutBuffers[0].cbBuffer = 0;

	OutBuffer.cBuffers = 1;
	OutBuffer.pBuffers = OutBuffers;
	OutBuffer.ulVersion = SECBUFFER_VERSION;

	DWORD dwSSPIOutFlags;
	TimeStamp tsExpiry;
	Status = m_SecurityFunc.InitializeSecurityContext( phCreds, phContext, NULL, dwSSPIFlags, 0, /*SECURITY_NATIVE_DREP*/0, NULL, 0, phContext, &OutBuffer, &dwSSPIOutFlags, &tsExpiry );
	if( FAILED(Status) )
	{
	    m_SecurityFunc.DeleteSecurityContext(phContext);
	    return Status;
	}

	PBYTE pbMessage = (BYTE *)(OutBuffers[0].pvBuffer);
	DWORD cbMessage = OutBuffers[0].cbBuffer;

	if( pbMessage != NULL && cbMessage != 0 )
	{
		m_bAllowPlainText = TRUE;
		m_bConInit = FALSE;
		cbData = Send( pbMessage, cbMessage );
		m_bAllowPlainText = FALSE;
		if( cbData == SOCKET_ERROR || cbData == 0 )
		{
			m_SecurityFunc.DeleteSecurityContext(phContext);
		    return WSAGetLastError();
		}

		m_SecurityFunc.FreeContextBuffer(pbMessage);
	}

    m_SecurityFunc.DeleteSecurityContext(phContext);
    return Status;
}

LONG CSecureSyncSocketEx::ServerDisconect( PCredHandle phCreds, CtxtHandle *phContext )
{
    DWORD dwType = SCHANNEL_SHUTDOWN;

    SecBuffer OutBuffers[1];
    OutBuffers[0].pvBuffer = &dwType;
    OutBuffers[0].BufferType = SECBUFFER_TOKEN;
    OutBuffers[0].cbBuffer = sizeof(dwType);

    SecBufferDesc OutBuffer;
    OutBuffer.cBuffers = 1;
    OutBuffer.pBuffers = OutBuffers;
    OutBuffer.ulVersion = SECBUFFER_VERSION;

	DWORD Status = m_SecurityFunc.ApplyControlToken( phContext, &OutBuffer );
	if( FAILED(Status) )
	{
	    m_SecurityFunc.DeleteSecurityContext(phContext);
	    return Status;
	}

	DWORD dwSSPIFlags = ASC_REQ_SEQUENCE_DETECT | ASC_REQ_REPLAY_DETECT | ASC_REQ_CONFIDENTIALITY | ASC_REQ_EXTENDED_ERROR | ASC_REQ_ALLOCATE_MEMORY | ASC_REQ_STREAM;
    DWORD dwSSPIOutFlags;

	OutBuffers[0].pvBuffer = NULL;
	OutBuffers[0].BufferType = SECBUFFER_TOKEN;
	OutBuffers[0].cbBuffer = 0;

	OutBuffer.cBuffers = 1;
	OutBuffer.pBuffers = OutBuffers;
	OutBuffer.ulVersion = SECBUFFER_VERSION;

    TimeStamp tsExpiry;
	Status = m_SecurityFunc.AcceptSecurityContext( phCreds, phContext, NULL, dwSSPIFlags, SECURITY_NATIVE_DREP, NULL, &OutBuffer, &dwSSPIOutFlags, &tsExpiry );
	if( FAILED(Status) )
	{
	    m_SecurityFunc.DeleteSecurityContext(phContext);
	    return Status;
	}

	PBYTE pbMessage = (BYTE *)(OutBuffers[0].pvBuffer);
	DWORD cbMessage = OutBuffers[0].cbBuffer;
	if( pbMessage != NULL && cbMessage != 0 )
	{
		m_bAllowPlainText = TRUE;
		m_bConInit = FALSE;
		DWORD cbData = Send( pbMessage, cbMessage );
		m_bAllowPlainText = FALSE;
		if( cbData == SOCKET_ERROR || cbData == 0 )
		{
			m_SecurityFunc.DeleteSecurityContext(phContext);
			return WSAGetLastError();
		}

		m_SecurityFunc.FreeContextBuffer(pbMessage);
	}

    m_SecurityFunc.DeleteSecurityContext(phContext);
    return Status;
}

BOOL CSecureSyncSocketEx::ServerConnect(SOCKADDR* lpSockAddr, int* lpSockAddrLen)
{
	if( !ServerHandshakeLoop( &m_hContext, &m_hCreds, m_bAuthClient, TRUE, TRUE ) )
	{
		return FALSE;
	}

	if( m_bAuthClient )
	{
		PCCERT_CONTEXT pRemoteCertContext = NULL;
		if( m_SecurityFunc.QueryContextAttributes( &m_hContext, SECPKG_ATTR_REMOTE_CERT_CONTEXT, (PVOID)&pRemoteCertContext ) != SEC_E_OK )
		{
			return FALSE;
		}
		else
		{
			if( ServerVerifyCertificate( pRemoteCertContext, 0 ) )
			{
				return FALSE;
			}
			::CertFreeCertificateContext( pRemoteCertContext );
		}
	}

	SecPkgContext_StreamSizes Sizes;
	if( m_SecurityFunc.QueryContextAttributes( &m_hContext, SECPKG_ATTR_STREAM_SIZES, &Sizes ) != SEC_E_OK )
	{
		return FALSE;
	}

	return TRUE;
}

BOOL CSecureSyncSocketEx::ServerHandshakeLoop( PCtxtHandle phContext, PCredHandle phCreds, BOOL fClientAuth, BOOL fDoInitialRead, BOOL NewContext )
{
	TimeStamp tsExpiry;
	SECURITY_STATUS scRet = SEC_E_SECPKG_NOT_FOUND;
	SecBufferDesc InBuffer;
	SecBufferDesc OutBuffer;
	SecBuffer InBuffers[2];
	SecBuffer OutBuffers[1];
	DWORD err = 0;

	BOOL fDoRead = fDoInitialRead;
	BOOL fInitContext = NewContext;

	DWORD dwSSPIFlags = ASC_REQ_SEQUENCE_DETECT | ASC_REQ_REPLAY_DETECT | ASC_REQ_CONFIDENTIALITY | ASC_REQ_EXTENDED_ERROR | ASC_REQ_ALLOCATE_MEMORY | ASC_REQ_STREAM;
	DWORD dwSSPIOutFlags;

	BYTE IoBuffer[IO_BUFFER_SIZE];
	DWORD cbIoBuffer = 0;

	if( fClientAuth )
	{
		dwSSPIFlags |= ASC_REQ_MUTUAL_AUTH;
	}

    OutBuffer.cBuffers = 1;
    OutBuffer.pBuffers = OutBuffers;
    OutBuffer.ulVersion = SECBUFFER_VERSION;

    scRet = SEC_I_CONTINUE_NEEDED;

    while( scRet == SEC_I_CONTINUE_NEEDED || scRet == SEC_E_INCOMPLETE_MESSAGE || scRet == SEC_I_INCOMPLETE_CREDENTIALS )
	{
        if( 0 == cbIoBuffer || scRet == SEC_E_INCOMPLETE_MESSAGE )
		{
            if( fDoRead )
			{
				m_bAllowPlainText = TRUE;
				err = Receive( IoBuffer+cbIoBuffer, IO_BUFFER_SIZE );
				m_bAllowPlainText = FALSE;

				if( err == SOCKET_ERROR || err == 0 )
				{
					return FALSE;
				}
				else
				{
					cbIoBuffer += err;
				}
            }
			else
			{
                fDoRead = TRUE;
            }
        }

        InBuffers[0].pvBuffer = IoBuffer;
        InBuffers[0].cbBuffer = cbIoBuffer;
        InBuffers[0].BufferType = SECBUFFER_TOKEN;

        InBuffers[1].pvBuffer = NULL;
        InBuffers[1].cbBuffer = 0;
        InBuffers[1].BufferType = SECBUFFER_EMPTY;

        InBuffer.cBuffers = 2;
        InBuffer.pBuffers = InBuffers;
        InBuffer.ulVersion = SECBUFFER_VERSION;

        OutBuffers[0].pvBuffer = NULL;
        OutBuffers[0].BufferType = SECBUFFER_TOKEN;
        OutBuffers[0].cbBuffer = 0;

        scRet = m_SecurityFunc.AcceptSecurityContext( phCreds, ( fInitContext ? NULL : phContext ), &InBuffer, dwSSPIFlags, SECURITY_NATIVE_DREP, ( fInitContext ? phContext : NULL ), &OutBuffer, &dwSSPIOutFlags, &tsExpiry );
        fInitContext = FALSE;

        if( scRet == SEC_E_OK || scRet == SEC_I_CONTINUE_NEEDED || ( FAILED(scRet) && ( 0 != ( dwSSPIOutFlags & ISC_RET_EXTENDED_ERROR ) ) ) )
		{
			if( OutBuffers[0].cbBuffer != 0 && OutBuffers[0].pvBuffer != NULL )
			{
				m_bAllowPlainText = TRUE;
                err = Send( OutBuffers[0].pvBuffer,OutBuffers[0].cbBuffer );
				m_bAllowPlainText = FALSE;

                m_SecurityFunc.FreeContextBuffer( OutBuffers[0].pvBuffer );
                OutBuffers[0].pvBuffer = NULL;
            }
        }

		if( scRet == SEC_E_OK )
		{
			if( InBuffers[1].BufferType == SECBUFFER_EXTRA )
			{
				::memcpy( IoBuffer, (LPBYTE)( IoBuffer + ( cbIoBuffer - InBuffers[1].cbBuffer ) ), InBuffers[1].cbBuffer );
				cbIoBuffer = InBuffers[1].cbBuffer;
            }
			else
			{
				cbIoBuffer = 0;
			}

			return TRUE;
		}
		else if( FAILED(scRet) && scRet != SEC_E_INCOMPLETE_MESSAGE )
		{
			return FALSE;
        }

		if( scRet != SEC_E_INCOMPLETE_MESSAGE && scRet != SEC_I_INCOMPLETE_CREDENTIALS )
		{
			if( InBuffers[1].BufferType == SECBUFFER_EXTRA )
			{
				::memcpy( IoBuffer, (LPBYTE)( IoBuffer + ( cbIoBuffer - InBuffers[1].cbBuffer ) ), InBuffers[1].cbBuffer );
				cbIoBuffer = InBuffers[1].cbBuffer;
			}
			else
			{
				cbIoBuffer = 0;
			}
		}
	}

    return FALSE;
}

DWORD CSecureSyncSocketEx::ServerVerifyCertificate( PCCERT_CONTEXT pServerCert, DWORD dwCertFlags )
{
	if( pServerCert == NULL )
	{
		return (DWORD)SEC_E_WRONG_PRINCIPAL;
	}

	int iRc = ::CertVerifyTimeValidity( NULL, pServerCert->pCertInfo );
	if( iRc != 0 )
	{
		return (DWORD)SEC_E_CERT_EXPIRED;
	}

	CERT_CHAIN_PARA ChainPara;
	::memset( &ChainPara, 0, sizeof( ChainPara ) );
	ChainPara.cbSize = sizeof( ChainPara );

	PCCERT_CHAIN_CONTEXT pChainContext = NULL;
	if( !::CertGetCertificateChain( NULL, pServerCert, NULL, NULL, &ChainPara, CERT_CHAIN_REVOCATION_CHECK_CHAIN, NULL, &pChainContext ) )
	{
		return ::GetLastError();
	}

	HTTPSPolicyCallbackData polHttps;
	::memset( &polHttps, 0, sizeof( HTTPSPolicyCallbackData ) );
	polHttps.cbStruct = sizeof( HTTPSPolicyCallbackData );
	polHttps.dwAuthType = AUTHTYPE_CLIENT;
	polHttps.fdwChecks = dwCertFlags;
	polHttps.pwszServerName = NULL;

	CERT_CHAIN_POLICY_PARA PolicyPara;
	::memset( &PolicyPara, 0, sizeof( PolicyPara ) );
	PolicyPara.cbSize = sizeof( PolicyPara );
	PolicyPara.pvExtraPolicyPara = &polHttps;

	CERT_CHAIN_POLICY_STATUS PolicyStatus;
	::memset( &PolicyStatus, 0, sizeof( PolicyStatus ) );
	PolicyStatus.cbSize = sizeof( PolicyStatus );

	if( !::CertVerifyCertificateChainPolicy( CERT_CHAIN_POLICY_SSL, pChainContext, &PolicyPara, &PolicyStatus ) )
	{
	    if( pChainContext )
		{
			::CertFreeCertificateChain(pChainContext);
		}
		return ::GetLastError();
	}

	if( PolicyStatus.dwError )
	{
	    if( pChainContext )
		{
			::CertFreeCertificateChain(pChainContext);
		}
		return PolicyStatus.dwError;
	}

	PCERT_CONTEXT *pCerts = new PCERT_CONTEXT[pChainContext->cChain];
	for( DWORD i = 0; i < pChainContext->cChain; i++ )
	{
		pCerts[i] = (PCERT_CONTEXT)(pChainContext->rgpChain[i]->rgpElement[0]->pCertContext);
	}
		
	CERT_REVOCATION_STATUS revStat;
	revStat.cbSize = sizeof( CERT_REVOCATION_STATUS );

	if( !::CertVerifyRevocation( X509_ASN_ENCODING, CERT_CONTEXT_REVOCATION_TYPE, pChainContext->cChain, (void **)pCerts, CERT_VERIFY_REV_CHAIN_FLAG, NULL, &revStat ) )
	{
	    if( pChainContext )
		{
			::CertFreeCertificateChain(pChainContext);
		}
		return ::GetLastError();
	}

	delete [] pCerts;

    if( pChainContext )
	{
		::CertFreeCertificateChain(pChainContext);
    }

    return SEC_E_OK;
}
