/////////////////////////////////////////////////////////////////////////////
// Implementation for a MFC class encapsulation of the SSL/TLS sockets

/*
 * Copyright (c) 2003 - 2005 by Daniel Weber (daniel@astolat.de)
 * All rights reserved.
 *
 * This is a SSL/TLS-implementation for Windows-Sockets in C++ written by
 * Daniel Weber (daniel@astolat.de) based an examples provided by
 * Martin Ziacek (martin.ziacek@blueyonder.co.uk).
 *
 * You are allowed to use, modify and redistribute this source-code for
 * commercial and non-commercial software as long as the following
 * conditions are met:
 * - Copyright remains bei Daniel Weber and redistributions of the source-
 *   code must retain this copyright-notice, list of conditions and
 *   disclaimer.
 * - Redistributions in binary form must mention Daniel Weber as author
 *   of parts of the SSL/TLS-implementations at program startup and/or
 *   in the documentation.
 *
 * This source-code is provided by Daniel Weber "as is" and any express or
 * implied warranties, including but not limited to the implied warranties
 * of merchantability and fitness for a particular purpose are disclaimed.
 * The authors or contributors are not liable for any damages arising out
 * of the use of this source-code, even if advised of the possibility of
 * such damage.
 */

class CSecureSyncSocketEx : public CSyncSocketEx
{
	DECLARE_DYNAMIC(CSecureSyncSocketEx);

// Construction
public:
	CSecureSyncSocketEx();
//	CSecureSyncSocketEx( const CSecureSyncSocketEx& rSrc );
//	void operator=( const CSecureSyncSocketEx& rSrc );
	BOOL Create( UINT nSocketPort = 0, int nSocketType = SOCK_STREAM, LPCTSTR lpszSocketAddress = NULL, BOOL bUseSecurity = FALSE, LPCTSTR pszCertName = NULL, BOOL bMachineStore = FALSE, DWORD dwProtocol = 0 );
	BOOL LateHandshake( BOOL bAuthClient = FALSE, LPCTSTR pszCertName = NULL, BOOL bMachineStore = FALSE, DWORD dwProtocol = 0, LPCTSTR pszHostName = NULL );
	BOOL IsSecured();

// Attributes
private:
	BOOL LoadSecurityLibrary();

	SECURITY_STATUS ClientCreateCredentials( LPCTSTR pszUserName, PCredHandle phCreds );
	BOOL ClientConnect( LPCTSTR szHostName );
	LONG ClientDisconnect( PCredHandle phCreds, CtxtHandle *phContext );
	SECURITY_STATUS ClientHandshake( PCredHandle phCreds, LPCTSTR pszServerName, CtxtHandle *phContext, SecBuffer *pExtraData );
	SECURITY_STATUS ClientHandshakeLoop( PCredHandle phCreds, CtxtHandle *phContext, BOOL fDoInitialRead, SecBuffer *pExtraData );
	DWORD ClientVerifyCertificate( PCCERT_CONTEXT pServerCert, LPCTSTR pszServerName, DWORD dwCertFlags );

	SECURITY_STATUS ServerCreateCredentials( LPCTSTR pszUserName, PCredHandle phCreds );
	BOOL ServerConnect( SOCKADDR* lpSockAddr, int* lpSockAddrLen );
	LONG ServerDisconect( PCredHandle phCreds, CtxtHandle *phContext );
	BOOL ServerHandshakeLoop( PCtxtHandle phContext, PCredHandle phCreds, BOOL fClientAuth, BOOL fDoInitialRead, BOOL NewContext );
	DWORD ServerVerifyCertificate( PCCERT_CONTEXT  pServerCert, DWORD dwCertFlags );

protected:
	BOOL m_bUseSecurity;
	BOOL m_bServer;

	CString m_strCertName;
	BOOL m_bMachineStore;
	DWORD m_dwProtocol;

	PCCERT_CONTEXT m_pCertContext;

	BOOL m_bAuthClient;

	HCERTSTORE m_hMyCertStore;
	SCHANNEL_CRED m_SchannelCred;
	CredHandle m_hCreds;
	CtxtHandle m_hContext;

	HMODULE m_hSecurity;
	SecurityFunctionTable m_SecurityFunc;

	BOOL m_bConInit;
	BOOL m_bAllowPlainText;

	BYTE *m_pbReceiveBuf;
	DWORD m_dwReceiveBuf;

	BYTE *m_pbIoBuffer;
	DWORD m_cbIoBuffer;

// Operations
public:
	virtual BOOL Accept( CSecureSyncSocketEx& rConnectedSocket, SOCKADDR* lpSockAddr = NULL, int* lpSockAddrLen = NULL );

	virtual BOOL Close();

	virtual BOOL Connect( LPCTSTR lpszHostAddress, UINT nHostPort );
	BOOL Connect( const SOCKADDR* lpSockAddr, int nSockAddrLen );

	BOOL Listen( int nConnectionBacklog = 5, BOOL bAuthClient = FALSE );

	virtual int Receive( void* lpBuf, int nBufLen, int nFlags = 0 );

	virtual int Send( const void* lpBuf, int nBufLen, int nFlags = 0 );
	virtual int Send( LPCTSTR lpszBuf );

// Implementation
public:
	virtual ~CSecureSyncSocketEx();

/*
	static CSyncSocketEx* PASCAL LookupHandle( SOCKET hSocket, BOOL bDead = FALSE );
	static void PASCAL AttachHandle( SOCKET hSocket, CSyncSocketEx* pSocket, BOOL bDead = FALSE );
	static void PASCAL DetachHandle( SOCKET hSocket, BOOL bDead = FALSE );
	static void PASCAL KillSocket( SOCKET hSocket, CSyncSocketEx* pSocket );
	static void PASCAL DoCallBack( WPARAM wParam, LPARAM lParam );

#ifdef _DEBUG
	virtual void AssertValid() const;
	virtual void Dump(CDumpContext& dc) const;
#endif

protected:
	virtual BOOL ConnectHelper( const SOCKADDR* lpSockAddr, int nSockAddrLen );
	virtual int ReceiveFromHelper( void* lpBuf, int nBufLen, SOCKADDR* lpSockAddr, int* lpSockAddrLen, int nFlags );
	virtual int SendToHelper( const void* lpBuf, int nBufLen, const SOCKADDR* lpSockAddr, int nSockAddrLen, int nFlags );
*/
};

inline int CSecureSyncSocketEx::Send( LPCTSTR lpszBuf )
{
	return Send( lpszBuf, _tcsclen( lpszBuf ) );
}

inline BOOL CSecureSyncSocketEx::IsSecured()
{
	return m_bUseSecurity && m_bConInit;
}
